package de.hftstuttgart.dtabackend.rest.v1.task;

import de.hftstuttgart.dtabackend.utils.*;
import de.hftstuttgart.dtabackend.models.ResultSummary;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.tika.Tika;
import org.springframework.core.env.Environment;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jakarta.servlet.annotation.MultipartConfig;

/**
 * Rest controller for handling code assignment file uploads and testing.
 */
@RestController
@RequestMapping("/v1/task/*")
@MultipartConfig
public class TaskUpload {

    private static final Logger LOG = LogManager.getLogger(TaskUpload.class);

    private final RepoUtil repoUtil;
    private final Path testTmpPath;
    private final ExecuteTestUtil executeTestUtil;
    private final Tika tika;

    public TaskUpload(Environment env, RepoUtil repoUtil, ExecuteTestUtil executeTestUtil) {
        this.repoUtil = repoUtil;
        this.executeTestUtil = executeTestUtil;
        this.testTmpPath = Paths.get(env.getProperty("tests.tmp.dir"));
        this.tika = new Tika();
    }

    @RequestMapping(method = RequestMethod.POST)
    public ResultSummary uploadAndTestFile(@RequestParam("taskFile") MultipartFile taskFileRef,
                                           @RequestParam("assignmentId") String assignmentId)
            throws IOException, InterruptedException {
        LOG.info("Submission for testing received");

        Path workDirectory = createWorkDirectory();
        Path srcPath = defineSourcePath(workDirectory);

        String mimeType = tika.detect(taskFileRef.getInputStream());
        processUploadedFile(taskFileRef, srcPath, mimeType);

        ResultSummary resultSummary = executeTestUtil.runTests(assignmentId, workDirectory);

        if (isPlainTextFile(mimeType)) {
            processTicketingInformation(taskFileRef, resultSummary);
        }

        LOG.info("Submission tested successfully");
        return resultSummary;
    }

    private Path createWorkDirectory() throws IOException {
        LOG.debug("Creating new temporary directory");
        Path workDirectory = Files.createTempDirectory(testTmpPath, "dta-submission");
        LOG.debug("Working directory for test: {}", workDirectory.toAbsolutePath());
        return workDirectory;
    }

    private Path defineSourcePath(Path workDirectory) {
        Path srcPath = workDirectory.resolve("src");
        LOG.debug("Source path defined as: {}", srcPath.toAbsolutePath());
        return srcPath;
    }

    private void processUploadedFile(MultipartFile taskFileRef, Path srcPath, String mimeType) throws IOException {
        switch (mimeType) {
            case "text/plain":
                handlePlainTextFile(taskFileRef, srcPath);
                break;

            case "application/zip":
                handleZipFile(taskFileRef, srcPath);
                break;

            default:
                handleUnsupportedFileType(mimeType);
        }
    }

    private void handlePlainTextFile(MultipartFile taskFileRef, Path srcPath) throws IOException {
        LOG.debug("Text file uploaded, searching for DTA config");

        Matcher config = findRepositoryConfig(taskFileRef);
        String subDir = config != null ? config.group(4) : "";

        LOG.debug("Cloning repository");
        repoUtil.cloneRepository(config, srcPath.toAbsolutePath().toString(), subDir);
    }

    private Matcher findRepositoryConfig(MultipartFile taskFileRef) throws IOException {
        Matcher config = RegexUtil.extractConfig(taskFileRef.getInputStream(), Pattern.compile(RegexUtil.DTA_SUBMISSIONCONFIGREGEX));

        if (config == null) {
            config = RegexUtil.extractConfig(taskFileRef.getInputStream(), Pattern.compile(RegexUtil.SUBMISSIONCONFIGREGEX));

            if (config == null) {
                throw new RuntimeException("Couldn't find repo config for student submission clone");
            }
        }
        return config;
    }

    private void handleZipFile(MultipartFile taskFileRef, Path srcPath) throws IOException {
        LOG.debug("ZIP archive uploaded, extracting content");
        ArchiveUtil.extractProjectFromZip(taskFileRef.getInputStream(), srcPath.toAbsolutePath());
    }

    private void handleUnsupportedFileType(String mimeType) {
        String msg = String.format("Couldn't process uploaded file with MIME type: %s", mimeType);
        LOG.error(msg);
        throw new RuntimeException(msg);
    }

    private boolean isPlainTextFile(String mimeType) {
        return "text/plain".equals(mimeType);
    }

    private void processTicketingInformation(MultipartFile taskFileRef, ResultSummary resultSummary) throws IOException {
        LOG.info("Checking for provided Ticketing system information");
        UnifiedTicketingUtil.reportResults(taskFileRef.getInputStream(), resultSummary);
    }
}
